#include "critbit.h"

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

struct CritBitNode {
    unsigned int isleaf:1;

    union {
        struct {
            unsigned int byte:31;
            uint8_t others;
            struct CritBitNode *left; /* lefts are 0s */
            struct CritBitNode *right; /* rights are 1s */
        };
        struct {
            unsigned int valuelen:31;
            void *value;
            char *key;
        };
    };
};

void crit_bit_init(struct CritBitTree *tree, void (*free_data)(void *data, int datalen)) {
    if (!tree) return;

    tree->count = 0;
    tree->root = 0;
    tree->free_data = free_data;
}

static void free_node(void (*free_data)(void *data, int datalen), struct CritBitNode *node) {
    if (!node) return;
    if (node->isleaf) {
        if (free_data)
            free_data(node->value, node->valuelen);
        free(node->key);
        free(node);
    } else {
        free_node(free_data, node->left);
        free_node(free_data, node->right);
        free(node);
    }
}

void crit_bit_free(struct CritBitTree *tree) {
    if (tree->root)
        free_node(tree->free_data, tree->root);

    tree->root = 0;
    tree->count = 0;
}

static struct CritBitNode *makeleaf(const char *key, size_t keylen, void *data, int datalen) {
    struct CritBitNode *leaf = malloc(sizeof *leaf);
    if (!leaf)
        return 0;

    leaf->key = malloc(keylen + 1);
    if (!leaf->key) {
        free(leaf);
        return 0;
    }

    memcpy(leaf->key, key, keylen + 1);
    leaf->value = data;
    leaf->valuelen = datalen;
    leaf->isleaf = 1;

    return leaf;
}

int crit_bit_add(struct CritBitTree *tree, const char *key, void *data, int datalen) {
    if (!tree) return 0;

    const uint8_t *keybytes = (void*)key;
    size_t keylen = strlen(key);
    struct CritBitNode *node = tree->root;

    if (!node) {
        struct CritBitNode *root = malloc(sizeof *root);
        root->key = malloc(keylen + 1);
        strcpy(root->key, key);
        root->valuelen = datalen;
        root->value = data;
        root->isleaf = 1;
        tree->root = root;
        tree->count++;
        return 0;
    }

    while (node && !node->isleaf) {
        uint8_t check = 0;
        if (node->byte < keylen)
            check = keybytes[node->byte];
        if ((node->others | check) == 0xFF)
            node = node->right;
        else
            node = node->left;
    }

    if (!node)
        return -1;

    uint32_t newbyte;
    uint32_t newothers = ~0; /* sigil for finding existing keys */
    for (newbyte = 0; newbyte < keylen; ++newbyte) {
        if (node->key[newbyte] != key[newbyte]) {
            newothers = node->key[newbyte] ^ key[newbyte];
            break;
        }
    }

    if (newothers == ~0) {
        if (node->key[newbyte] != 0) {
            newothers = node->key[newbyte];
        }
    }

    if (newothers == ~0) {
        if (tree->free_data)
            tree->free_data(node->value, node->valuelen);
        node->value = data;
        node->valuelen = datalen;
        return 0;
    }

    /* bit hacks */
    while (newothers & (newothers - 1))
        newothers &= newothers - 1;
    newothers ^= 0xFF;
    uint8_t c = node->key[newbyte];
    int isright = (1 + (newothers | c)) >> 8;

    struct CritBitNode *newnode = malloc(sizeof *newnode);
    if (!newnode)
        return -2;

    struct CritBitNode *leafnode = makeleaf(key, keylen, data, datalen);
    if (!leafnode)
        return -2;

    newnode->isleaf = 0;
    newnode->byte = newbyte;
    newnode->others = newothers;

    if (isright) {
        newnode->left = leafnode;
    } else {
        newnode->right = leafnode;
    }

    struct CritBitNode **parent = (struct CritBitNode **)&tree->root;
    for (;;) {
        struct CritBitNode *here = *parent;
        if (here->isleaf) break;
        if (here->byte > newbyte) break;
        if (here->byte == newbyte && here->others > newothers) break;

        uint8_t c2 = 0;
        if (here->byte < keylen) c2 = keybytes[here->byte];
        if ((1 + (here->others | c2)) > 0xFF) {
            parent = &here->right;
        } else {
            parent = &here->left;
        }
    }

    if (isright) {
        newnode->right = *parent;
    } else {
        newnode->left = *parent;
    }
    *parent = newnode;
    tree->count++;
    return 0;
}

int crit_bit_has(const struct CritBitTree *tree, const char *key) {
    if (!tree) return 0;

    const uint8_t *keybytes = (void*)key;
    size_t keylen = strlen(key);
    const struct CritBitNode *node = tree->root;

    while (node && !node->isleaf) {
        uint8_t check = 0;
        if (node->byte < keylen)
            check = keybytes[node->byte];
        if ((node->others | check) == 0xFF)
            node = node->right;
        else
            node = node->left;
    }

    return node && !strcmp(key, node->key);
}

void *crit_bit_get(const struct CritBitTree *tree, const char *key, int *len) {
    if (!tree) return 0;

    const uint8_t *keybytes = (void*)key;
    size_t keylen = strlen(key);
    const struct CritBitNode *node = tree->root;

    while (node && !node->isleaf) {
        uint8_t check = 0;
        if (node->byte < keylen)
            check = keybytes[node->byte];
        if ((node->others | check) == 0xFF)
            node = node->right;
        else
            node = node->left;
    }

    if (!node) return 0;

    if (len)
        *len = node->valuelen;
    return node->value;
}

int crit_bit_del(struct CritBitTree *tree, const char *key) {
    if (!tree) return 1;

    const uint8_t *keybytes = (void*)key;
    size_t keylen = strlen(key);
    struct CritBitNode *node = tree->root;
    struct CritBitNode *prev = 0;
    void *tmp = &(tree->root);
    struct CritBitNode **parent = (struct CritBitNode **)tmp;
    struct CritBitNode **grandparent = 0;
    int isright;

    if (!node) return 1;

    while (node && !node->isleaf) {
        grandparent = parent;
        uint8_t c = 0;
        if (node->byte < keylen) c = keybytes[node->byte];
        isright = ((1 + (node->others | c)) > 0xFF);
        if (isright) {
            parent = &node->right;
        } else {
            parent = &node->left;
        }
        prev = node;
        node = *parent;
    }

    if (!node)
        return 1;

    if (strcmp(key, node->key))
        return 1;

    if (tree->free_data) tree->free_data(node->value, node->valuelen);
    free(node->key);
    free(node);
    tree->count--;

    if (!grandparent || !prev) {
        tree->root = 0;
        return 0;
    }

    if (isright) {
        *grandparent = prev->left;
    } else {
        *grandparent = prev->right;
    }

    free(prev);
    return 0;
}

int crit_bit_size(const struct CritBitTree *tree) {
    return tree->count;
}

static int iterate_node(int (*callback)(const char*,void*,int,void*), void *data, struct CritBitNode *node) {
    if (!node) return 0;
    if (node->isleaf) {
        return callback(node->key, node->value, node->valuelen, data);
    } else {
        int r;
        r = iterate_node(callback, data, node->left);
        if (r) return r;
        r = iterate_node(callback, data, node->right);
        return r;
    }
}

int crit_bit_foreach(const struct CritBitTree *tree, int (*callback)(const char*,void*,int,void*), void *data) {
    if (!tree) return 0;
    if (!tree->root) return 0;
    if (!callback) return 0;

    return iterate_node(callback, data, tree->root);
}
