#include <stdlib.h>
#include <string.h>

#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <argp.h>

#include "jit.h"
#include "args.h"
#include "architectures.h"

char *read_dynamic(int fd, int *len) {
    char buf[0x1000];
    ssize_t n;
    char *result = 0;
    *len = 0;
    while ((n = read(fd, buf, sizeof buf))) {
        if (n < 0) {
            if (errno == EAGAIN)
                continue;
            perror("read");
            break;
        }

        char *nresult = realloc(result, *len + n + 1);
        if (!nresult) {
            free(result);
            return 0;
        }
        result = nresult;
        memcpy(result + *len, buf, n);
        *len += n;
        result[*len] = 0;
    }

    return result;
}

char *read_filename(const char *fname, int *len, int *mapped) {
    if (fname[0] == '-' && fname[1] == 0) {
        *mapped = 0;
        return read_dynamic(0, len);
    } else {
        char *result = 0;
        int fd = open(fname, O_RDONLY);
        *len = lseek(fd, 0, SEEK_END);
        if (*len < 0) {
            /* fallback to dynamic memory */
            *len = 0;
            *mapped = 0;
            result = read_dynamic(fd, len);
        } else {
            *mapped = 1;
            result = mmap(0, *len, PROT_READ, MAP_PRIVATE, fd, 0);
        }
        close(fd);
        return result;
    }
}

void free_read_file(char *contents, int len, int mapped) {
    if (!mapped) {
        free(contents);
    } else {
        munmap(contents, len);
    }
}

void dump_str(const char *str, void *unused) {
    puts(str);
}

int main(int argc, char *argv[]) {
    struct Arguments args = { 0 };
    int offset = parse_arguments(&args, argc, argv);

    int len;
    int mapped;
    char *source = read_filename(args.program, &len, &mapped);
    int result;

    switch (args.behaviour) {
        case BEHAVIOUR_INTERPRET: {
            jit interp = get_implementation(ARCH_INTERPRET);
            if (!interp) {
                free_read_file(source, len, mapped);
                return -1;
            }

            main_func interp_main = interp(source, len);
            free_read_file(source, len, mapped);
            if (!interp_main) return -1;

            result = interp_main(argc - offset, argv + offset);
            break;
        }

        case BEHAVIOUR_JIT: {
            jit jitter = get_implementation(args.jit.arch);
            if (!jitter) {
                free_read_file(source, len, mapped);
                return -1;
            }

            main_func jitter_main = jitter(source, len);
            free_read_file(source, len, mapped);
            result = jitter_main(argc - offset, argv + offset);
            break;
        }

        case BEHAVIOUR_ASM:
            asm_implementation(args.assemble.arch, source, len, dump_str, 0);
            free_read_file(source, len, mapped);
            result = 0;
            break;
    }

    return result;
}
